package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
import it.unimi.dsi.util.Properties;
import it.unimi.dsi.util.StringMap;

import java.io.IOException;
import java.io.InvalidClassException;

import org.apache.commons.configuration.ConfigurationException;

/** Static container providing commodity methods to load archives.
 * 
 * <p>This class contains methods to load easily {@link Archive} instances whose implementing
 * classes adopt the following suggestion: they implement a
 * <code>getInstance()</code> method accepting a basename (a {@link CharSequence}), a
 * {@linkplain Properties set of properties} and an optional {@link StringMap} (acting as a URI list). 
 * 
 * <p>The file <var><samp>basename</samp></var> + {@link #PROPERTIES_EXTENSION} should contain
 * the properties of the archive, which will be passed on to the <code>getInstance()</code> method
 * (possibly after some massaging). The property file <strong>must</strong>
 * contain a property {@link PropertyKeys#CLASS}, so that we can dispatch the 
 * appropriate <code>getInstance()</code> method by reflection.
 * 
 * <p>An optional boolean parameter may be used to specify that an archive must be loaded entirely in core memory:
 * in this case, this class will add the key {@link PropertyKeys#INMEMORY} 
 * add to the archive property set. The key can also be set manually in the property file if necessary.
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 * 
 */
public class ArchiveLoader {

	/** The (capitalized) names for archive properties. */
	public static enum PropertyKeys {
		/** The name of the class that should load the archive. */
		CLASS,
		/** The number of documents in the archive. */
		DOCUMENTS,
		/** The number of terms in the archive. */
		TERMS,
		/** The number of words in the archive (i.e., the sum of the lengths of all documents). */
		WORDS,
		/** Specify whether the archive is to be loaded in core memory. */
		INMEMORY,
	}

	/** The extension of a properties file. */
	public static final String PROPERTIES_EXTENSION = ".properties";

	private ArchiveLoader() {}

	/**
	 * Creates an archive instance using the given basename (and no URI list).
	 * 
	 * @param basename the basename of the archive.
	 * @param inMemory whether the archive should be loaded in core memory.
	 * @return an archive instance.
	 */
	public static Archive<? extends DocumentSummary> getInstance( final CharSequence basename, final boolean inMemory ) throws IOException, ConfigurationException {
		return getInstance( basename, null, inMemory );
	}

	/**
	 * Creates an archive instance using the given basename and loading the given URI list.
	 * 
	 * @param basename the basename of the archive.
	 * @param uriList the filename of a URI list.
	 * @param inMemory whether the archive should be loaded in core memory.
	 * @return an archive instance.
	 */
	@SuppressWarnings("unchecked")
	public static Archive<? extends DocumentSummary> getInstance( final CharSequence basename, final CharSequence uriList, final boolean inMemory ) throws IOException, ConfigurationException {
		final Properties properties = new Properties( basename + PROPERTIES_EXTENSION );
		if ( inMemory ) properties.setProperty( PropertyKeys.INMEMORY, true );

		if ( !properties.containsKey( PropertyKeys.CLASS ) ) throw new InvalidClassException( "No " + PropertyKeys.CLASS.toString().toLowerCase() + " property found in " + basename + PROPERTIES_EXTENSION );

		try {
			return (Archive<? extends DocumentSummary>)Class.forName( properties.getString( PropertyKeys.CLASS ) ).getMethod( "getInstance", CharSequence.class, Properties.class, CharSequence.class )
					.invoke( null, basename, properties, uriList );
		}
		catch ( Exception e ) {
			throw new RuntimeException( e );
		}
	}

	/**
	 * Creates a disk-based archive instance using the given basename (and no URI list).
	 * 
	 * @param basename the basename of the archive.
	 * @return an archive instance.
	 */
	public static Archive<? extends DocumentSummary> getInstance( final CharSequence basename ) throws IOException, ConfigurationException {
		return getInstance( basename, false );
	}

	/**
	 * Creates a disk-based archive instance using the given basename and loading the given URI list.
	 * 
	 * @param basename the basename of the archive.
	 * @param uriList the filename of a URI list.
	 * @return an archive instance.
	 */
	public static Archive<? extends DocumentSummary> getInstance( final CharSequence basename, final CharSequence uriList ) throws IOException, ConfigurationException {
		return getInstance( basename, uriList, false );
	}
}
