package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.law.vector.ImmutableSparseVector;
import cern.colt.Sorting;


/**
 * A featurizer implementing the BM25 weighting scheme.
 * 
 * @author Alessio Orlandi
 */
public class BM25Featurizer<T extends DocumentSummary> extends AbstractFeaturizer<T, ImmutableSparseVector> {
	private static final long serialVersionUID = 1993L;

	/** Average document length in the archive. */
	protected final double averageDocumentLength;

	/**
	 * Number of features to maintain. Only <code>cutAt</code>-top scored ones are returned. If
	 * set to {@link Integer#MAX_VALUE}, all features are maintained.
	 */
	protected final int cutAt;

	/** The IDF part of the BM25 weighting scheme, which depends on the document frequency, cached. */
	protected final double[] idfPart;

	/** Global number of terms in the archive. */
	protected final int totalTerms;

	private static final double K1 = 1.2, B = 0.5;

	private static final double EPSILON_SCORE = 0.0000001;

	public BM25Featurizer( final Archive<? extends T> archive, final int cutAt ) {
		super( archive );

		this.cutAt = cutAt;
		this.averageDocumentLength = (double)archive.numberOfWords() / archive.numberOfDocuments();
		this.idfPart = new double[ ( this.totalTerms = archive.numberOfTerms() ) ];

		for ( int i = 0; i < idfPart.length; i++ ) {
			final double freq = archive.frequency( i );
			idfPart[ i ] = Math.log( ( numberOfDocuments - freq + .5 ) / ( freq + .5 ) );
		}


	}

	public BM25Featurizer( final Archive<? extends T> archive ) {
		this( archive, Integer.MAX_VALUE );
	}

	@SuppressWarnings("unchecked")
	public ImmutableSparseVector get( final Object o ) {
		final T summary = (T)o;

		final int nTerms = summary.size();

		double[] scores = new double[ nTerms ];
		final double[] scores2 = new double[ nTerms ];
		int[] terms = new int[ nTerms ];


		final int totalOccourrences = summary.length();
		for ( int i = 0; i < nTerms; i++ ) {
			final int term = summary.term( i );
			final int count = summary.count( i );

			terms[ i ] = term;
			scores[ i ] = scores2[ i ] = ( Math.max( EPSILON_SCORE, idfPart[ term ] ) * ( K1 + 1.0 ) * count ) / ( count + K1 * ( ( 1 - B + B * totalOccourrences ) / averageDocumentLength ) );


		}

		double threshold = 0.0;

		if ( cutAt < nTerms && cutAt < Integer.MAX_VALUE ) {
			Sorting.mergeSort( scores2, 0, scores2.length );
			threshold = scores2[ nTerms - cutAt + 1 ];
		}

		return ImmutableSparseVector.getInstance( terms, scores, totalTerms, threshold, summary.id() );
	}

}
