package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import java.net.URI;



/**
 * A summary of the term information of a document.
 * 
 * <p>Summaries contain a document id which is unique within the archive, and optionally a
 * {@link URI} providing the original document. Beside that, they provide the document
 * {@linkplain #length() length} (e.g. before term pruning), the {@linkplain #term(int) terms}
 * appearing in the document and their {@linkplain #count(int) counts} (a.k.a. within-document
 * frequencies). Document lengths and counts might not be available, depending on the 
 * archive, and in this case special values are returned.
 * 
 * <p>Additionally, if {@link #sorted()} returns true then terms appear in increasing order. That
 * is, <code>term(i) &lt; term(j)</code> for <code>i &lt; j</code>. You can force this property
 * by calling {@link #sort()}.
 * 
 * <p>The equality contract for document summaries is that summaries containing the same
 * information should be equal, independently of the order in which it is provided. Note that you
 * should not use {@link #sort()} when implementing {@link Object#equals(Object)}, as the user does
 * not expect the object to change when it is compared.
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */
public interface DocumentSummary {

	/**
	 * Returns the id of the document this summary represents.
	 * 
	 * @return the id of the document this summary represents.
	 */
	public int id();

	/**
	 * Returns the number of terms in this summary.
	 * 
	 * <p>Note that due to pruning (e.g., of <i>hapax legomena</i>) the number of terms might be
	 * different from the number of terms of the document this summary represents.
	 * 
	 * @return the number of terms in this summary.
	 */
	public int size();

	/**
	 * Returns the term of given index.
	 * 
	 * @return the term of given index.
	 */
	public int term( int index );

	/**
	 * Returns the count (a.k.a. within-document frequency) of the term of given index, or -1 if
	 * counts are not available.
	 * 
	 * @return the count of the term of given index, or -1 if counts are not available.
	 */
	public int count( int index );

	/**
	 * Returns the index of the given term.
	 * 
	 * @param term a term number.
	 * @return the index of the given term in this summary, or -1 if the term does not appear in
	 * this summary.
	 */
	public int indexOf( int term );

	/**
	 * A URI representing the source of this summary.
	 * 
	 * @return a URI representing the source of this summary, or <code>null</code> if no such URI
	 * is available.
	 */
	public URI uri();

	/**
	 * Returns the length in words of the document this summary represents, or -1 if lengths are not available.
	 * 
	 * @return the length in words of the document this summary represents, or -1 if lengths are not available.
	 */
	public int length();

	/**
	 * Returns whether this summary is sorted.
	 * 
	 * @return whether this summary is sorted.
	 */
	public boolean sorted();

	/**
	 * Sorts this summary is sorted.
	 * 
	 * <p>After calling this method, {@link #sorted()} will return true.
	 * 
	 * @return this document summary.
	 */
	public DocumentSummary sort();
}
