package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.law.vector.ImmutableSparseVector;
import cern.colt.Sorting;

/**
 * A featurizer that returns TF-IDF featurizer.
 * 
 * @author Alessio Orlandi
 */
public class TFIDFFeaturizer<T extends DocumentSummary> extends AbstractFeaturizer<T, ImmutableSparseVector> {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1992L;

	/**
	 * The IDF part of the weighting scheme, which depends on the document frequency, cached.
	 */
	protected final double[] idfPart;

	private final int totalTerms;

	private int cutAt;


	public TFIDFFeaturizer( final Archive<? extends T> archive, int cutAt ) {
		super( archive );

		this.idfPart = new double[ ( this.totalTerms = archive.numberOfTerms() ) ];
		this.cutAt = cutAt;

		for ( int i = 0; i < totalTerms; i++ ) {
			final int f = archive.frequency( i );
			if ( f > 0 ) idfPart[ i ] = Math.log( (double)numberOfDocuments / (double)f );
		}


	}

	public TFIDFFeaturizer( final Archive<? extends T> archive ) {
		this( archive, Integer.MAX_VALUE );
	}

	@SuppressWarnings("unchecked")
	public ImmutableSparseVector get( final Object o ) {
		final T summary = (T)o;

		final int nTerms = summary.size();

		double[] scores = new double[ nTerms ];
		final double[] scores2 = new double[ nTerms ];
		int[] terms = new int[ nTerms ];


		final int totalOccourrences = summary.length();
		for ( int i = 0; i < nTerms; i++ ) {
			final int term = summary.term( i );
			final int count = summary.count( i );

			terms[ i ] = term;
			scores[ i ] = scores2[ i ] = (double)count / (double)totalOccourrences * idfPart[ term ];
		}

		double threshold = 0.0;

		if ( cutAt > nTerms && cutAt < Integer.MAX_VALUE ) {
			Sorting.mergeSort( scores2, 0, scores2.length );
			threshold = scores2[ scores2.length - cutAt + 1 ];
		}

		return ImmutableSparseVector.getInstance( terms, scores, totalTerms, threshold, summary.id() );
	}


}
