package it.unimi.di.archive4j.scratch;


/*
 * Copyright (C) 2006-2013 Paolo Boldi, Massimo Santini and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
import it.unimi.di.archive4j.ArchiveLoader;
import it.unimi.di.archive4j.ArrayDocumentSummary;
import it.unimi.di.archive4j.SequentialBitstreamArchive;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Coding;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.io.NullOutputStream;
import it.unimi.dsi.io.OutputBitStream;
import it.unimi.dsi.logging.ProgressLogger;

import java.io.IOException;
import java.io.PrintWriter;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cern.colt.GenericSorting;
import cern.colt.Swapper;
import cern.colt.function.IntComparator;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.UnflaggedOption;

/**
 * Estimates position cost.
 * 
 * @author Alessio Orlandi
 */

public class EstimatePositionsLocal {
	protected static final Logger LOGGER = LoggerFactory.getLogger( EstimatePositionsLocal.class );


	static NullOutputStream nos = NullOutputStream.getInstance();

	static OutputBitStream obs = new OutputBitStream( nos, 0 );

	static protected Coding coding;

	static long[] counters;

	static long written = 0;

	static IntArrayList bitlengths = new IntArrayList();

	public static void processSummary( final ArrayDocumentSummary s ) throws IOException {

		/* Add fictional element. */
		int size = s.size();
		int filled = 0;


		final int[] tpTerms = new int[ size + 1 ];
		final int[] tpCounts = new int[ size + 1 ];

		for ( int i = 0; i < size; i++ ) {
			tpTerms[ i ] = s.term( i );
			tpCounts[ i ] = s.count( i );
			filled += tpCounts[ i ];
		}

		tpTerms[ size ] = -1;
		tpCounts[ size ] = s.length() - filled;

		GenericSorting.quickSort( 0, tpTerms.length, new IntComparator() {
			public int compare( final int a, final int b ) {
				return tpCounts[ b ] - tpCounts[ a ];
			}
		}, new Swapper() {
			public void swap( final int a, final int b ) {
				int tmp = tpTerms[ a ];
				tpTerms[ a ] = tpTerms[ b ];
				tpTerms[ b ] = tmp;
				tmp = tpCounts[ a ];
				tpCounts[ a ] = tpCounts[ b ];
				tpCounts[ b ] = tmp;
			}
		} );

		long writtenBefore = written;
		for ( int i = 0; i < tpTerms.length; i++ ) {
			// Each position is represented by the term rank in the sorted list.
			obs.writtenBits( 0 );

			switch ( coding ) {
			case DELTA:
				obs.writeDelta( i );
				break;
			case GAMMA:
				obs.writeGamma( i );
				break;
			case NIBBLE:
				obs.writeNibble( i );
				break;
			case UNARY: // FIXME This is an hack.
				obs.writeMinimalBinary( i, size + 1 );
				break;
			case ZETA_2:
				obs.writeZeta( i, 2 );
				break;
			case ZETA_3:
				obs.writeZeta( i, 3 );
				break;
			case ZETA_4:
				obs.writeZeta( i, 4 );
				break;
			case ZETA_5:
				obs.writeZeta( i, 5 );
				break;
			case SHIFTED_GAMMA:
				obs.writeShiftedGamma( i );
				break;

			default:
				throw new RuntimeException( "Invalid value!" );
			}

			final int quantity = (int)obs.writtenBits() * tpCounts[ i ];
			counters[ tpTerms[ i ] < 0 ? counters.length - 1 : tpTerms[ i ] ] += quantity;
			written += quantity;
		}

		bitlengths.add( (int)( written - writtenBefore ) );
	}


	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {

		SimpleJSAP jsap = new SimpleJSAP( EstimatePositions.class.getName(), "Estimates.", new Parameter[] {
				new FlaggedOption( "coding", JSAP.STRING_PARSER, "UNARY", JSAP.NOT_REQUIRED, 'c', "coding", "Specify code -- use UNARY for Minimal Binary" ),
				new FlaggedOption( "archive", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'A', "archive", "Base archive" ),
				new UnflaggedOption( "output", JSAP.STRING_PARSER, JSAP.REQUIRED, "The filename for the output  data file" )

		} );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		ProgressLogger pl = new ProgressLogger( LOGGER );


		SequentialBitstreamArchive archive = (SequentialBitstreamArchive)ArchiveLoader.getInstance( jsapResult.getString( "archive" ) );

		counters = new long[ archive.numberOfTerms() + 1 ];

		coding = Coding.valueOf( jsapResult.getString( "coding" ) );

		pl.itemsName = "documents";
		pl.logInterval = ProgressLogger.TEN_SECONDS;
		pl.displayFreeMemory = true;
		pl.expectedUpdates = archive.numberOfDocuments();
		pl.start( "Reading documents..." );


		for ( ArrayDocumentSummary summary : archive ) {
			processSummary( summary );
			pl.lightUpdate();
		}

		pl.done();
		archive.close();

		PrintWriter pw = new PrintWriter( jsapResult.getString( "output" ) + ".byterms.txt" );

		pw.println( "Total number of written bytes: " + written / 8 );
		pw.println( "Instances for the 'missing' term " + counters[ counters.length - 1 ] );
		pw.println( "Counters for single term bits " );
		pw.println( "----" );
		for ( int i = 0; i < counters.length - 1; i++ )
			pw.println( i + " " + counters[ i ] );
		pw.close();

		pw = new PrintWriter( jsapResult.getString( "output" ) + ".bydoc.txt" );
		pw.println( "Bitlength of each document" );
		for ( int len : bitlengths )
			pw.println( len );
		pw.close();
	}


}
