package it.unimi.di.archive4j.tool;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.big.mg4j.document.AbstractDocumentSequence;
import it.unimi.di.big.mg4j.document.DocumentFactory;
import it.unimi.di.big.mg4j.document.DocumentSequence;
import it.unimi.di.big.mg4j.document.IdentityDocumentFactory;
import it.unimi.di.big.mg4j.document.InputStreamDocumentSequence;
import it.unimi.di.big.mg4j.document.PropertyBasedDocumentFactory;
import it.unimi.di.big.mg4j.index.NullTermProcessor;
import it.unimi.di.big.mg4j.index.TermProcessor;
import it.unimi.di.big.mg4j.util.MG4JClassParser;
import it.unimi.di.archive4j.ArrayDocumentSummary;
import it.unimi.di.archive4j.BitstreamArchiveWriter;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Coding;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Component;
import it.unimi.di.archive4j.tool.MergePreprocessedData.FrequencyFilter;
import it.unimi.di.archive4j.tool.MergePreprocessedData.LengthFilter;
import it.unimi.di.archive4j.tool.MergePreprocessedData.MixedFilter;
import it.unimi.di.archive4j.tool.MergePreprocessedData.StopwordFilter;
import it.unimi.di.archive4j.tool.MergePreprocessedData.TermFilter;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.io.FastBufferedInputStream;
import it.unimi.dsi.fastutil.io.FastBufferedOutputStream;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.io.FileLinesCollection;
import it.unimi.dsi.io.InputBitStream;
import it.unimi.dsi.lang.ObjectParser;
import it.unimi.dsi.util.StringMap;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import org.apache.commons.io.input.TeeInputStream;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;


/** Builds an archive from a document collection.
 * 
 * <p>This commodity class provides a main method that will call in turn
 * {@link Preprocess}, {@link MergePreprocessedData} and {@link Scan}. In case no
 * sequence is specified, documents are read from standard input. However, be warned that since
 * archive construction requires two passes, standard input will be copied to a temporary file. 
 * 
 * @author Sebastiano Vigna
 */

public class ArchiveBuilder {
	private final static Logger LOGGER = LoggerFactory.getLogger( ArchiveBuilder.class );
	
	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {

		SimpleJSAP jsap = new SimpleJSAP( ArchiveBuilder.class.getName(), "Builds an archive out of a given document sequence; if no sequence is provided, a list of documents separated by a delimiter will be read from standard input.",
				new Parameter[] {

			new FlaggedOption( "sequence", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'S', "sequence", "A serialised document sequence." ),
			new FlaggedOption( "objectSequence", new ObjectParser( DocumentSequence.class, MG4JClassParser.PACKAGE ), JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'o', "object-sequence", "An object specification describing a document sequence." ),
			new FlaggedOption( "indexedField", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, 'I', "indexed-field", "The field of the document factory that will be indexed." ),
			new FlaggedOption( "termProcessor", new ObjectParser( TermProcessor.class, MG4JClassParser.PACKAGE, new String[] { "getInstance" } ),	NullTermProcessor.class.getName(), JSAP.NOT_REQUIRED, 't', "termprocessor", "The term pocessor to be used." ),
			new FlaggedOption( "delimiter", JSAP.INTEGER_PARSER, Integer.toString( it.unimi.di.big.mg4j.tool.Scan.DEFAULT_DELIMITER ), JSAP.NOT_REQUIRED, 'd', "delimiter", "The document delimiter." ),
			new FlaggedOption( "factory", MG4JClassParser.getParser(), IdentityDocumentFactory.class.getName(), JSAP.NOT_REQUIRED, 'f', "factory", "A document factory with a standard constructor." ),
			new FlaggedOption( "property", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'p', "property", "A 'key=value' specification, or the name of a property file" ).setAllowMultipleDeclarations( true ),
			new Switch( "remove-mixed", 'x', "remove-mixed", "Remove mixed alphanumeric terms." ),
			new FlaggedOption( "stopwords", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 's', "stopwords", "List of lowercase, termprocessed stopwords to be eliminated, one per line." ),
			new FlaggedOption( "min-length", JSAP.INTEGER_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.NO_SHORTFLAG, "min-length", "Minimum length of a term." ),
			new FlaggedOption( "max-length", JSAP.INTEGER_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.NO_SHORTFLAG, "max-length", "Maximum length of a term." ),
			new FlaggedOption( "min-freq", JSAP.INTEGER_PARSER, "0", JSAP.NOT_REQUIRED, 'm', "min-freq", "Minimum frequency for a term." ),
			new FlaggedOption( "max-freq", JSAP.INTEGER_PARSER, "-1", JSAP.NOT_REQUIRED, 'M', "max-freq", "Maximum frequency of a term." ),
			new FlaggedOption( "coding", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'c', "coding", "Specify code (" + StringUtils.join( Coding.values(), ", " ) + ") for a given component (" + StringUtils.join( Component.values(), ", " ) + ") in the form COMPONENT:CODING." ).setAllowMultipleDeclarations( true ),
			new Switch( "randomAccess", 'r', "random-access", "Build a random-access archive (input documents must be in increasing order)." ),
			new FlaggedOption( "uriMap", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'u', "uriMap", "A map from URIs to document identifiers that will be used to renumber documents." ),
			new UnflaggedOption( "basename", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The archive output basename." ),
		} );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		final String basename = jsapResult.getString( "basename" );
		
		if ( jsapResult.userSpecified( "sequence" ) && jsapResult.userSpecified( "objectSequence" ) ) throw new IllegalArgumentException( "You cannot specify both a serialised and an parseable-object sequence" );

		DocumentSequence sequence;
		DocumentFactory factory = null;
		File inputDuplicate = null;
		int delimiter = -1;
		FastBufferedOutputStream inputDuplicateStream = null;
		if ( jsapResult.userSpecified( "objectSequence" ) ) sequence = (DocumentSequence)jsapResult.getObject( "objectSequence" );
		else if ( jsapResult.userSpecified( "sequence" ) ) sequence = AbstractDocumentSequence.load( jsapResult.getString( "sequence" ) );
		else {
			delimiter = jsapResult.getInt( "delimiter" );
			LOGGER.debug( "Documents will be separated by the Unicode character " + delimiter );
			factory = PropertyBasedDocumentFactory.getInstance( jsapResult.getClass( "factory" ), jsapResult.getStringArray( "property" ) );
			inputDuplicate = File.createTempFile( ArchiveBuilder.class.getName(), "input" );
			inputDuplicate.deleteOnExit();
			sequence = new InputStreamDocumentSequence( new TeeInputStream( System.in, inputDuplicateStream = new FastBufferedOutputStream( new FileOutputStream( inputDuplicate ) ) ), delimiter, factory );
		}
		
		
		final String indexedField = jsapResult.getString( "indexedField" );
		final TermProcessor termProcessor = (TermProcessor)jsapResult.getObject( "termProcessor" );
		
		Preprocess.run( basename, sequence, termProcessor,  jsapResult.getString( "indexedField" ) );
		sequence.close();
		if ( inputDuplicateStream != null ) inputDuplicateStream.close();

		ObjectArrayList<TermFilter> killers = new ObjectArrayList<TermFilter>();

		if ( jsapResult.userSpecified( "stopwords" ) ) killers.add( new StopwordFilter( new FileLinesCollection( jsapResult.getString( "stopwords" ), "UTF-8" ).allLines() ) );

		if ( jsapResult.getBoolean( "remove-mixed" ) ) killers.add( new MixedFilter() );

		if ( jsapResult.userSpecified( "min-freq" ) || jsapResult.userSpecified( "max-freq" ) ) killers
				.add( new FrequencyFilter( jsapResult.getInt( "min-freq" ), jsapResult.getInt( "max-freq" ) ) );

		if ( jsapResult.userSpecified( "max-length" ) ) killers.add( new LengthFilter( jsapResult.getInt( "min-length" ), jsapResult.getInt( "max-length" ) ) );

		MergePreprocessedData.run( basename, jsapResult.getString( "basename" ), killers.toArray( new TermFilter[ 0 ] ) );
		
		StringMap<? extends CharSequence> terms = (StringMap<? extends CharSequence>)BinIO.loadObject( basename + MergePreprocessedData.TERMMAP_EXTENSION );

		int[] frequency = new int[ terms.size() ];
		final InputBitStream frequencies = new InputBitStream( basename + Preprocess.FREQUENCIES_EXTENSION );
		frequencies.readGammas( frequency, terms.size() );
		frequencies.close();
		
		final int[][] sortingPermutations = BitstreamArchiveWriter.makeSortingPermutations( frequency );
		frequency = null; // Free memory

		final StringMap urls = jsapResult.userSpecified( "uriMap" ) ? (StringMap<CharSequence>)BinIO.loadObject( jsapResult.getString( "uriMap" ) ) : null;

		final BitstreamArchiveWriter<ArrayDocumentSummary> writer = ( new BitstreamArchiveWriter<ArrayDocumentSummary>( jsapResult.getString( "basename" ),
				sortingPermutations, jsapResult.getBoolean( "randomAccess" ), CompressionFlags.valueOf( jsapResult.getStringArray( "coding" ), CompressionFlags.DEFAULT ) ) );

		// We need to re-open the document sequence.
		if ( jsapResult.userSpecified( "objectSequence" ) ) sequence = (DocumentSequence)jsapResult.getObject( "objectSequence" );
		else if ( jsapResult.userSpecified( "sequence" ) ) sequence = AbstractDocumentSequence.load( jsapResult.getString( "sequence" ) );
		else sequence = new InputStreamDocumentSequence( new FastBufferedInputStream( new FileInputStream( inputDuplicate ) ), delimiter, factory );

		Scan.run( sequence, termProcessor, terms, urls, writer, indexedField );
		
		sequence.close();
		if ( inputDuplicate != null ) inputDuplicate.delete();
	}
}
