package it.unimi.di.archive4j.tool;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.archive4j.Archive;
import it.unimi.di.archive4j.ArchiveLoader;
import it.unimi.di.archive4j.DocumentSummary;
import it.unimi.dsi.io.OutputBitStream;
import it.unimi.dsi.logging.ProgressLogger;

import java.io.IOException;
import java.util.Iterator;

import org.apache.commons.configuration.ConfigurationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.UnflaggedOption;

/** Analyses an {@link Archive} and generates frequency and global count data.
 * 
 * @author Sebastiano Vigna
 */

public class ComputeArchiveData {
	private static final Logger LOGGER = LoggerFactory.getLogger( ComputeArchiveData.class );

	/** Computes frequency and count data for the given archive.
	 * 
	 * @param basename the archive basename.
	 */
	public static void run( final String basename ) throws IOException, ConfigurationException {
		final Archive<? extends DocumentSummary> archive = ArchiveLoader.getInstance( basename );
		run( archive, basename );
		archive.close();
	}

	/** Computes frequency and count data for the given archive, and saves the result using the given basename.
	 * 
	 * @param archive an archive.
	 * @param basename the output basename.
	 */
	public static void run( final Archive<? extends DocumentSummary> archive, final String basename ) throws IOException {

		final Iterator<? extends DocumentSummary> iterator = archive.iterator();
		final ProgressLogger pl = new ProgressLogger( LOGGER );

		pl.expectedUpdates = archive.numberOfDocuments();
		pl.start( "Scanning archive..." );
		final int numberOfTerms = archive.numberOfTerms();
		final int[] frequency = new int[ numberOfTerms ];
		final long[] count = new long[ numberOfTerms ];

		DocumentSummary s;
		int t;
		while( iterator.hasNext() ) {
			s = iterator.next();
			for( int i = s.size(); i-- != 0; ) {
				t = s.term( i );
				frequency[ t ]++;
				if ( s.count( i ) != -1 ) count[ t ] += s.count( i );
			}
			
			pl.lightUpdate();
		}

		pl.done();
		
		final OutputBitStream frequencies = new OutputBitStream( basename + Preprocess.FREQUENCIES_EXTENSION );
		for( int i = 0; i < numberOfTerms; i++ ) frequencies.writeGamma( frequency[ i ] );
		frequencies.close();
		
		final OutputBitStream counts = new OutputBitStream( basename + Preprocess.COUNTS_EXTENSION );
		for( int i = 0; i < numberOfTerms; i++ ) counts.writeLongGamma( count[ i ] );
		counts.close();
	}


	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {
		final SimpleJSAP jsap = new SimpleJSAP( ComputeArchiveData.class.getName(), "Computes frequencies and global counts of an archive, and stores them using the standard SequentialBitStreamArchive extensions.", 
			new Parameter[] {
				new UnflaggedOption( "archive", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The archive basename." ),
				new UnflaggedOption( "basename", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The output basename." ) 
			}
		);

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		final Archive archive = ArchiveLoader.getInstance( jsapResult.getString( "archive" ) ); 
		run( archive, jsapResult.getString( "basename" ) );
		archive.close();
	}
}
