package it.unimi.di.archive4j.tool;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.archive4j.Archive;
import it.unimi.di.archive4j.ArchiveLoader;
import it.unimi.di.archive4j.ArchiveWriter;
import it.unimi.di.archive4j.BitstreamArchiveWriter;
import it.unimi.di.archive4j.DocumentSummary;
import it.unimi.di.archive4j.SequentialBitstreamArchive;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Coding;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Component;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.fastutil.objects.ObjectHeapSemiIndirectPriorityQueue;
import it.unimi.dsi.logging.ProgressLogger;

import java.io.IOException;
import java.net.URI;
import java.util.Comparator;
import java.util.Iterator;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;


/** Merges or concatenates sorted archives.
 * 
 * <p><em>Concatenating</em> involves renumbering the documents appearing in each archive (the first summary
 * of the second archive is offset using the id of the last summary of the first archive plus one, and so on).
 * <em>Merging</em> assumes that no document id appears twice&mdash;no renumbering is performed.
 * 
 * @author Alessio Orlandi
 * 
 */
public class MergeSortedArchives {
	private static final Logger LOGGER = LoggerFactory.getLogger( MergeSortedArchives.class );

	/** A wrapper returning the document summaries of an underlying iterator with their id suitably offset. */
	private static class ConcatenationIterator extends AbstractObjectIterator<DocumentSummary> {
		final Iterator<? extends DocumentSummary> delegateIterator;
		final int offset;
		
		public ConcatenationIterator( Iterator<? extends DocumentSummary> delegate, int offset ) {
			this.delegateIterator = delegate;
			this.offset = offset;
		}
		
		public boolean hasNext() {
			return delegateIterator.hasNext();
		}
		
		public DocumentSummary next() {
			final DocumentSummary delegate = delegateIterator.next();
			return new DocumentSummary() {
				
				public int count( int index ) {
					return delegate.count( index );
				}

				public int id() {
					return delegate.id() + offset;
				}

				public int indexOf( int term ) {
					return delegate.indexOf( term );
				}

				public int length() {
					return delegate.length();
				}

				public int size() {
					return delegate.size();
				}

				public int term( int index ) {
					return delegate.term ( index );
				}

				public URI uri() {
					return delegate.uri();
				}

				public boolean sorted() {
					return delegate.sorted();
				}
				
				public DocumentSummary sort() {
					return delegate.sort();
				}
			};
		}
	}

	
	@SuppressWarnings("unchecked")
	public static void run( ArchiveWriter<DocumentSummary> output, Archive<? extends DocumentSummary>[] inputs, boolean concatenate, ProgressLogger pl ) throws IOException {
		final int n = inputs.length;
		final int[] prefixSum = new int[ n + 1 ]; // This is actually useful only for concatenation.
		
		int estimatedOutputSize = 0;
		for ( int i = 0; i < n; i++ ) {
			estimatedOutputSize += inputs[ i ].numberOfDocuments();
			if ( concatenate) prefixSum[ i + 1 ] = prefixSum[ i ] + inputs[ i ].getDocumentByIndex( inputs[ i ].numberOfDocuments() - 1 ).id() + 1;
		}

		pl.itemsName = "documents";
		pl.expectedUpdates = estimatedOutputSize;
		pl.start( ( concatenate ? "Concatenating " : "Merging " ) + n + " archives (" + estimatedOutputSize + " documents)..." );

		final Iterator<? extends DocumentSummary> iterator[] = new Iterator[ n ];
		final DocumentSummary document[] = new DocumentSummary[ n ];
		final ObjectHeapSemiIndirectPriorityQueue<DocumentSummary> queue = new ObjectHeapSemiIndirectPriorityQueue<DocumentSummary>( document, n, new Comparator<DocumentSummary>() {
			public int compare( final DocumentSummary x, final DocumentSummary y ) {
				return x.id() - y.id();
			}
		} );

		for ( int i = 0; i < n; i++ ) {
			iterator[ i ] = ( concatenate ? new ConcatenationIterator ( inputs[ i ].iterator(), prefixSum[i] ) : inputs[ i ].iterator() ) ;
			if ( iterator[ i ].hasNext() ) {
				document[ i ] = iterator[ i ].next();
				queue.enqueue( i );
			}
		}
		queue.allChanged();

		int first;
		while ( !queue.isEmpty() ) {
			final DocumentSummary d = document[ first = queue.first() ];

			if ( LOGGER.isDebugEnabled() ) LOGGER.debug( "Appending document " + d.id() );
			output.append( d );

			if ( iterator[ first ].hasNext() ) {
				document[ first ] = iterator[ first ].next();
				queue.changed();
			}
			else queue.dequeue();

			// Chase all documents with the same id.
			if ( !queue.isEmpty() && document[ ( first = queue.first() ) ].id() == d.id() )
			{
				final IllegalStateException e = new IllegalStateException( "Concatenation is not enabled and document ID " + d.id() + " is duplicate.");
				LOGGER.error(e.getMessage(), e );
				throw e;
			}
				
			pl.lightUpdate();
		}

		pl.done();

		for ( Archive<? extends DocumentSummary> a : inputs ) a.close();
		output.close();
	}


	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {

		SimpleJSAP jsap = new SimpleJSAP( MergeSortedArchives.class.getName(), "Merges sorted archives.", 
				new Parameter[] {
					new Switch( "concatenate", 'C', "concatenate", "Concatenate archives by renumbering documents."),
					new FlaggedOption( "coding", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'c', "coding", "Specify code (" + StringUtils.join( Coding.values(), ", " ) + ") for a given component (" + StringUtils.join( Component.values(), ", " ) + ") in the form COMPONENT:CODING." ).setAllowMultipleDeclarations( true ),
					new UnflaggedOption( "output", JSAP.STRING_PARSER, JSAP.REQUIRED, "Output basename." ),
					new UnflaggedOption( "input",  JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.GREEDY, "Input basenames." )
		});

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		final String[] inputNames = jsapResult.getStringArray( "input" );
		final SequentialBitstreamArchive[] inputs = new SequentialBitstreamArchive[ inputNames.length ];
		for ( int i = 0; i < inputNames.length; i++ ) {
			inputs[ i ] = (SequentialBitstreamArchive)ArchiveLoader.getInstance( inputNames[ i ] );
			if ( i > 0 && inputs[ i - 1 ].numberOfTerms() != inputs[ i ].numberOfTerms() ) throw new IllegalStateException( "All archives must have been generated using the same term map." );
		}	

		final BitstreamArchiveWriter<DocumentSummary> writer = new BitstreamArchiveWriter( jsapResult.getString( "output" ), inputs[0], true, CompressionFlags.valueOf( jsapResult.getStringArray( "coding" ), CompressionFlags.DEFAULT ) );

		run( writer, inputs, jsapResult.getBoolean( "concatenate" ), new ProgressLogger( LOGGER ) );
	}
}
