package it.unimi.di.archive4j.tool;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
import it.unimi.di.big.mg4j.document.Document;
import it.unimi.di.big.mg4j.document.DocumentCollection;
import it.unimi.di.big.mg4j.document.DocumentIterator;
import it.unimi.di.big.mg4j.document.DocumentSequence;
import it.unimi.di.big.mg4j.document.IdentityDocumentFactory;
import it.unimi.di.big.mg4j.index.TermProcessor;
import it.unimi.di.big.mg4j.util.MG4JClassParser;
import it.unimi.di.archive4j.ArrayDocumentSummary;
import it.unimi.di.archive4j.BitstreamArchiveWriter;
import it.unimi.di.archive4j.RandomAccessBitstreamArchive;
import it.unimi.di.archive4j.SequentialBitstreamArchive;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Coding;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Component;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.ints.Int2IntOpenHashMap;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.io.InputBitStream;
import it.unimi.dsi.io.WordReader;
import it.unimi.dsi.lang.MutableString;
import it.unimi.dsi.lang.ObjectParser;
import it.unimi.dsi.logging.ProgressLogger;
import it.unimi.dsi.util.Properties;
import it.unimi.dsi.util.StringMap;

import java.io.PrintWriter;
import java.io.Reader;
import java.net.URI;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;


/**
 * Scans a {@link DocumentSequence} to build a
 * {@link SequentialBitstreamArchive} or a {@link RandomAccessBitstreamArchive}, using preprocessed data
 * generated by {@link Preprocess} and {@link MergePreprocessedData}.
 * 
 * <p>Each document returned by the sequence is parsed, its terms are fed into a
 * {@linkplain TermProcessor term processor} and then checked against the term map provided by the
 * {@linkplain Preprocess preprocessing phase}. Those surviving the process are part of the
 * summary.
 * 
 * <p>An optional URI-to-id map can be provided to set the id of each document. The archive thus
 * obtained will not be in general sorted (but you can use {@link SortBitstreamArchive} to fix
 * that).
 * 
 * @see BitstreamArchiveWriter
 * @see SequentialBitstreamArchive
 * @see Preprocess
 * @see MergePreprocessedData
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */

public class Scan {

	private static final Logger LOGGER = LoggerFactory.getLogger( Scan.class );

	public static void run( DocumentSequence sequence, TermProcessor processor, StringMap<? extends CharSequence> terms, StringMap<? extends CharSequence> urls,
			BitstreamArchiveWriter<ArrayDocumentSummary> writer, String indexedField ) throws Exception {

		final MutableString word = new MutableString(), nonWord = new MutableString();
		final DocumentIterator iterator = sequence.iterator();

		final int field = sequence.factory().fieldIndex( indexedField );
		if ( field == -1 ) throw new IllegalArgumentException( "Unknown field " + indexedField );

		int count = 0;
		final PrintWriter outWriter = ( writer == null ? new PrintWriter( System.out ) : null );
		int numberOfTerms;
		int[] termCount = IntArrays.EMPTY_ARRAY, termsInDoc = IntArrays.EMPTY_ARRAY;
		int docLength, termNumber, localNumber;
		final ProgressLogger pl = new ProgressLogger( LOGGER );

		Int2IntOpenHashMap global2local = new Int2IntOpenHashMap( Hash.DEFAULT_INITIAL_SIZE, Hash.VERY_FAST_LOAD_FACTOR );
		global2local.defaultReturnValue( -1 );

		pl.itemsName = "documents";
		pl.logInterval = ProgressLogger.TEN_SECONDS;
		pl.displayFreeMemory = true;
		if ( sequence instanceof DocumentCollection ) pl.expectedUpdates = ( (DocumentCollection)sequence ).size();

		pl.start( "Scanning field " + indexedField + "..." );

		Document document;
		while ( ( document = iterator.nextDocument() ) != null ) {

			int docId = (int)( urls != null ? urls.getLong( document.uri() ) : count++ );
			if ( docId < 0 ) {
				LOGGER.info( "Skipping document with unknown id for uri : " + document.uri() );
				document.close();
				continue;
			}

			docLength = 0;
			numberOfTerms = 0;
			global2local.clear();

			final WordReader wordReader = document.wordReader( field );
			wordReader.setReader( (Reader)document.content( field ) );

			int len = 0;
			while ( wordReader.next( word, nonWord ) ) {
				if ( word.length() == 0 ) continue;
				len++;

				if ( !processor.processTerm( word ) ) continue;

				termNumber = (int)terms.getLong( word );

				if ( termNumber >= 0 ) {
					localNumber = global2local.get( termNumber );
					if ( localNumber == -1 ) {
						global2local.put( termNumber, localNumber = global2local.size() );
						if ( termsInDoc.length == numberOfTerms ) {
							termsInDoc = IntArrays.grow( termsInDoc, numberOfTerms + 1 );
							termCount = IntArrays.grow( termCount, numberOfTerms + 1 );
						}

						termCount[ numberOfTerms ] = 1;
						termsInDoc[ numberOfTerms++ ] = termNumber;
					}
					else termCount[ localNumber ]++;
				}
			}

			docLength += len;

			final ArrayDocumentSummary summary = new ArrayDocumentSummary( termsInDoc, termCount, numberOfTerms, docId, document.uri() == null ? null : URI.create( document.uri().toString() ), len );

			if ( writer != null ) writer.append( summary );
			else outWriter.println( summary );

			pl.lightUpdate();

			document.close();
		}

		if ( writer == null ) outWriter.close();
		else writer.close();

		pl.done();
		sequence.close();
	}


	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {

		SimpleJSAP jsap = new SimpleJSAP( Scan.class.getName(), "Builds an archive out of the given document sequence and preprocessed data.", new Parameter[] {
				new FlaggedOption( "sequence", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'S', "sequence", "A serialised document sequence that will be used instead of stdin." ),
				new FlaggedOption( "objectSequence", new ObjectParser( DocumentSequence.class, MG4JClassParser.PACKAGE ), JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'o', "object-sequence",
						"An object specification describing a document sequence that will be used instead of stdin." ),
				new FlaggedOption( "delimiter", JSAP.INTEGER_PARSER, Integer.toString( it.unimi.di.big.mg4j.tool.Scan.DEFAULT_DELIMITER ), JSAP.NOT_REQUIRED, 'd', "delimiter", "The document delimiter." ),
				new FlaggedOption( "factory", MG4JClassParser.getParser(), IdentityDocumentFactory.class.getName(), JSAP.NOT_REQUIRED, 'f', "factory",	"A document factory with a standard constructor." ),
				new FlaggedOption( "property", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'p', "property", "A 'key=value' specification, or the name of a property file." ).setAllowMultipleDeclarations( true ),
				new FlaggedOption( "indexedField", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, 'I', "indexed-field", "The field of the document factory that will be indexed." ),
				new FlaggedOption( "coding", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'c', "coding", "Specify code (" + StringUtils.join( Coding.values(), ", " )
						+ ") for a given component (" + StringUtils.join( Component.values(), ", " ) + ") in the form COMPONENT:CODING." ).setAllowMultipleDeclarations( true ),
				new Switch( "randomAccess", 'r', "random-access", "Build a random-access archive (input documents must be in increasing order)." ),
				new Switch( "ascii", 'A', "ascii", "Print ASCII (debugging) output on stdout." ),
				new FlaggedOption( "uriMap", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'u', "uriMap", "A map from URIs to document identifiers that will be used to renumber documents." ),
				new UnflaggedOption( "basename", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of the archive." ),
				new UnflaggedOption( "preprocessed", JSAP.STRING_PARSER, JSAP.NOT_REQUIRED, "The basename of preprocessed data. If missing, the archive basename will be used." ), } );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		if ( jsapResult.userSpecified( "sequence" ) && jsapResult.userSpecified( "objectSequence" ) ) throw new IllegalArgumentException(
				"You cannot specify both a serialised and a parseable-object sequence" );

		final String basename = jsapResult.getString( "basename" );
		final String preprocessed = jsapResult.userSpecified( "preprocessed" ) ? jsapResult.getString( "preprocessed" ) : basename;
		final Properties properties = new Properties( preprocessed + Preprocess.PROPERTIES_EXTENSION );
		
		int[] frequency = new int[ properties.getInt( MergePreprocessedData.PropertyKeys.TERMS )];
		new InputBitStream( preprocessed + Preprocess.FREQUENCIES_EXTENSION ).readGammas( frequency, frequency.length );	
		
		final int[][] sortingPermutations = BitstreamArchiveWriter.makeSortingPermutations( frequency );
		frequency = null; // Free memory
		
		final DocumentSequence sequence = jsapResult.userSpecified( "objectSequence" ) ? (DocumentSequence)jsapResult.getObject( "objectSequence" ) : it.unimi.di.big.mg4j.tool.Scan.getSequence( jsapResult
				.getString( "sequence" ), jsapResult.getClass( "factory" ), jsapResult.getStringArray( "property" ), jsapResult.getInt( "delimiter" ), LOGGER );
		final String indexedField = jsapResult.getString( "indexedField" );
		final StringMap urls = jsapResult.userSpecified( "uriMap" ) ? (StringMap<CharSequence>)BinIO.loadObject( jsapResult.getString( "uriMap" ) ) : null;

		if ( !properties.containsKey( MergePreprocessedData.PropertyKeys.TERMS ) )
			throw new IllegalArgumentException("Given preprocessing data requires merging.");
				
		final StringMap terms = (StringMap<CharSequence>)BinIO.loadObject( preprocessed + MergePreprocessedData.TERMMAP_EXTENSION );

		final boolean randomAccess = jsapResult.getBoolean("randomAccess") || urls == null;
		final BitstreamArchiveWriter<ArrayDocumentSummary> writer = jsapResult.getBoolean( "ascii" ) ? null : new BitstreamArchiveWriter<ArrayDocumentSummary>( basename, sortingPermutations,
				randomAccess, CompressionFlags.valueOf( jsapResult.getStringArray( "coding" ), CompressionFlags.DEFAULT ) );

		run( sequence, ObjectParser.fromSpec( properties.getString( Preprocess.PropertyKeys.TERMPROCESSOR ), TermProcessor.class, MG4JClassParser.PACKAGE, new String[] { "getInstance" } ), terms,
				urls, writer, indexedField );
	}
}
