package it.unimi.di.archive4j.tool;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.archive4j.ArchiveLoader;
import it.unimi.di.archive4j.ArrayDocumentSummary;
import it.unimi.di.archive4j.BitstreamArchiveWriter;
import it.unimi.di.archive4j.DocumentSummary;
import it.unimi.di.archive4j.RandomAccessBitstreamArchive;
import it.unimi.di.archive4j.SequentialBitstreamArchive;
import it.unimi.dsi.logging.ProgressLogger;

import java.io.File;
import java.io.IOException;
import java.util.Comparator;
import java.util.Iterator;

import org.apache.commons.configuration.ConfigurationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cern.colt.Sorting;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.UnflaggedOption;

/** Sorts bitstream archives. 
 * 
 * <p>{@linkplain SequentialBitstreamArchive Sequential archives} can be sorted
 * using this class so that they provide [{@linkplain RandomAccessBitstreamArchive random access}.
 * 
 * @author Alessio Orlandi
 */
public class SortBitstreamArchive {
	private static Logger LOGGER = LoggerFactory.getLogger( SortBitstreamArchive.class );

	public static String inMemorySize = "250000";

	/** Input and output basename for the archives. */
	private String inputBasename, outputBasename;

	/** Number of documents per blocks in memory sorting phase. */
	private int blockSize;

	/** Filenames of the sorted blocks temporary files. */
	private String[] sortedBlocks;

	/** The input bitstream archive. */
	private SequentialBitstreamArchive inputArchive;

	/**
	 * The two permutations used for the archive (the second being the one loaded from the input
	 * archive).
	 */
	private int[][] permutations;

	/** Empty constructor used for restarting. */
	public SortBitstreamArchive() {}

	/**
	 * First-start constructor.
	 * 
	 * @param inputBasename Unsorted archive basename
	 * @param outputBasename Sorted output archive basename
	 * @param blockSize Size, in documents, used for each memory-sorted block.
	 */
	public SortBitstreamArchive( final String inputBasename, final String outputBasename, final int blockSize ) throws IOException, ConfigurationException {

		this.inputBasename = inputBasename;
		this.outputBasename = outputBasename;

		inputArchive = (SequentialBitstreamArchive)ArchiveLoader.getInstance( inputBasename );

		permutations = new int[ 2 ][];

		int size;
		permutations[ 1 ] = inputArchive.getPermutation();
		permutations[ 0 ] = new int[ size = permutations[ 1 ].length ];
		for ( int i = 0; i < size; i++ )
			permutations[ 0 ][ permutations[ 1 ][ i ] ] = i;

		this.blockSize = blockSize;
		sortedBlocks = new String[ inputArchive.numberOfDocuments() ];


	}

	/** Runs the two phases. */
	public void run( ProgressLogger pl ) throws IOException {
		int ndocs = inputArchive.numberOfDocuments();
		int nblocks = ( ndocs / blockSize ) + ( ndocs % blockSize == 0 ? 0 : 1 );
		sortedBlocks = new String[ nblocks ];
		for ( int i = 0; i < sortedBlocks.length; i++ )
			sortedBlocks[ i ] = inputBasename + ".sortblock@" + i;

		innerSorting( pl );

		final RandomAccessBitstreamArchive[] blockArchives = new RandomAccessBitstreamArchive[ sortedBlocks.length ];
		try {
			for ( int i = 0; i < sortedBlocks.length; i++ )
				blockArchives[ i ] = (RandomAccessBitstreamArchive)ArchiveLoader.getInstance( sortedBlocks[ i ] );

			final BitstreamArchiveWriter<DocumentSummary> writer = new BitstreamArchiveWriter<DocumentSummary>( outputBasename,  permutations, true, inputArchive.getCodings() );
			MergeSortedArchives.run( writer, blockArchives, false, pl );
		}
		catch ( Exception e ) {
			throw new RuntimeException( e );
		}

		for ( int i = 0; i < sortedBlocks.length; i++ )
			new File( sortedBlocks[ i ] ).deleteOnExit();
	}




	/**
	 * Starts phase 1 (in memory sorting of blocks. )
	 * 
	 * @param pl a progress logger
	 * @throws IOException
	 */
	public void innerSorting( ProgressLogger pl ) throws IOException {

		final int ndocs = inputArchive.numberOfDocuments();
		final int nblocks = sortedBlocks.length;

		ArrayDocumentSummary[] inMemoryPhase = new ArrayDocumentSummary[ blockSize ];

		pl.itemsName = "blocks";
		pl.expectedUpdates = nblocks;
		pl.start( "In memory per-block sorting.." );

		Iterator<ArrayDocumentSummary> iterator = inputArchive.iterator();

		int currentBlock = 0;

		while ( currentBlock < nblocks ) {
			final int size = Math.min( blockSize, ndocs - currentBlock * blockSize );
			for ( int i = 0; i < size; i++ )
				inMemoryPhase[ i ] = iterator.next();

			Sorting.quickSort( inMemoryPhase, 0, size, new Comparator<DocumentSummary>() {

				public int compare( DocumentSummary x, DocumentSummary y ) {
					return x.id() - y.id();
				}

			} );
			LOGGER.debug( "Emitting block in " + sortedBlocks[ currentBlock ] );

			BitstreamArchiveWriter<ArrayDocumentSummary> writer = new BitstreamArchiveWriter<ArrayDocumentSummary>( sortedBlocks[ currentBlock ], permutations, true, 
					inputArchive.getCodings() );

			for ( int i = 0; i < size; i++ )
				writer.append( inMemoryPhase[ i ] );

			writer.close();

			currentBlock++;
			pl.update();
		}

		inMemoryPhase = null;
		pl.stop();
	}

	public static void main( String[] args ) throws Exception {

		SimpleJSAP jsap = new SimpleJSAP( SortBitstreamArchive.class.getName(), "Sorts a bitstream archive.", new Parameter[] {
				new FlaggedOption( "batchSize", JSAP.INTEGER_PARSER, inMemorySize, JSAP.NOT_REQUIRED, 'b', "batch-size", "Size in documents of a batch." ),
				new UnflaggedOption( "input", JSAP.STRING_PARSER, JSAP.REQUIRED, "Input basename." ),
				new UnflaggedOption( "output", JSAP.STRING_PARSER, JSAP.REQUIRED, "Output basename." ) 
		} );


		JSAPResult jsapResult = jsap.parse( args );

		if ( jsap.messagePrinted() ) return;
		final SortBitstreamArchive sorter = new SortBitstreamArchive( jsapResult.getString( "input" ), jsapResult.getString( "output" ), jsapResult.getInt( "batchSize" ) );

		sorter.run( new ProgressLogger( LOGGER ) );
	}
}
