package it.unimi.di.archive4j;

/*		 
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


import it.unimi.di.big.mg4j.document.AbstractDocument;
import it.unimi.di.big.mg4j.document.AbstractDocumentCollection;
import it.unimi.di.big.mg4j.document.Document;
import it.unimi.di.big.mg4j.document.DocumentCollection;
import it.unimi.di.big.mg4j.document.DocumentFactory;
import it.unimi.di.big.mg4j.document.IdentityDocumentFactory;
import it.unimi.di.big.mg4j.index.NullTermProcessor;
import it.unimi.di.big.mg4j.index.TermProcessor;
import it.unimi.di.archive4j.Archive;
import it.unimi.di.archive4j.ArchiveLoader;
import it.unimi.di.archive4j.ArrayArchive;
import it.unimi.di.archive4j.ArrayDocumentSummary;
import it.unimi.di.archive4j.BitstreamArchiveWriter;
import it.unimi.di.archive4j.DocumentSummary;
import it.unimi.di.archive4j.RandomAccessBitstreamArchive;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Coding;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Component;
import it.unimi.di.archive4j.tool.ComputeArchiveData;
import it.unimi.di.archive4j.tool.Preprocess;
import it.unimi.dsi.fastutil.ints.Int2IntAVLTreeMap;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.objects.Reference2ObjectMap;
import it.unimi.dsi.fastutil.objects.Reference2ObjectMaps;
import it.unimi.dsi.io.FastBufferedReader;
import it.unimi.dsi.io.WordReader;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.io.StringReader;
import java.net.URI;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;

import junit.framework.TestCase;


public class TestCollection {
	public static class InMemoryDocumentCollection extends AbstractDocumentCollection implements Serializable {
		private static final long serialVersionUID = 1L;

		protected final String[] docSources;

		protected final DocumentFactory factory = new IdentityDocumentFactory();

		protected final WordReader wordReader = new FastBufferedReader();

		public InMemoryDocumentCollection( String[] docSources ) {
			this.docSources = docSources;
		}

		@Override
		public void close() throws IOException {
			super.close();
		}

		public DocumentFactory factory() {
			return factory;
		}

		public DocumentCollection copy() {
			return new InMemoryDocumentCollection( docSources );
		}

		public Document document( final long idx ) throws IOException {
			return new AbstractDocument() {

				public Object content( int field ) throws IOException {
					return new StringReader( docSources[ (int)idx ] );
				}

				public CharSequence title() {
					return "Document " + idx;
				}

				public CharSequence uri() {
					return "http://document"+idx+"/";
				}

				public WordReader wordReader( int field ) {
					return wordReader;
				}
			};
		}

		@SuppressWarnings("unchecked")
		public Reference2ObjectMap<Enum<?>, Object> metadata( long document ) throws IOException {
			return Reference2ObjectMaps.EMPTY_MAP;
		}

		public long size() {
			return docSources.length;
		}

		public InputStream stream( long unused ) throws IOException {
			throw new UnsupportedOperationException();
		}
		
	 
	}



	final public static String[] testTerms = { "I", "am", "but", "can", "like", "not", "pretenD", "them" };

	// During the parsing phrase the term order is "I am not like them but can pretenD"
	final public static String[] testParseDocSources = { "I am not like them but I can pretenD", "like them but not like them", "pretenD them but not like I am", "I like but I am",
			"but I am but I am", "not", "" };

	final public static String[][] testBiParseDocSources = { new String[] { "I pretenD them", "but" }, new String[] { "like I am not", "I can" } };


	// Parsed version of testParseDocSources
	final public static String[] testParseDoc = { "0 9 8 0:2 1:1 2:1 3:1 4:1 5:1 6:1 7:1", "1 6 4 2:1 4:2 5:1 7:2", "2 7 7 0:1 1:1 2:1 4:1 5:1 6:1 7:1", "3 5 4 0:2 1:1 2:1 4:1", "4 6 3 0:2 1:2 2:2",
			"5 1 1 5:1", "6 0 0" };

	// Frequency for documents above should be: 4 4 5 1 4 4 2 3

	
	// On this version, we strip "I".
	final public static String[][] testBiParseDoc = { { "0 3 2 5:1 6:1", "1 1 1 1:1"} ,{ "0 4 3 0:1 3:1 4:1", "1 2 1 2:1" } };


	public static ArrayDocumentSummary[] testParse;

	public static ArrayDocumentSummary[][] testBiParse = new ArrayDocumentSummary[2][];

	// ID, wordlen, termlen, term1:count1 <= term2:count2 ...
	// These are based on testTerms but these docs are not for testing parsing methods, see
	// testParseDoc
	final public static String[] testDoc = { "0 12 3 0:1 3:3 4:1", "2 7 5 0:1 1:1 2:1 3:1 5:1", "4 9 1 6:99", "10 0 0", "11 100 2 2:3 7:1", "129 100 3 0:1 1:2 4:4" };

	final public static String[] extensions = new String[] { RandomAccessBitstreamArchive.MISSING_EXTENSION, Preprocess.FREQUENCIES_EXTENSION,
			RandomAccessBitstreamArchive.ARCHIVE_EXTENSION, RandomAccessBitstreamArchive.OFFSETS_EXTENSION, RandomAccessBitstreamArchive.PERM_EXTENSION, ArchiveLoader.PROPERTIES_EXTENSION,
			Preprocess.COUNTS_EXTENSION };

	final public static ArrayDocumentSummary[] testCollection1 = { ArrayDocumentSummary.parse( testDoc[ 4 ], URI.create( "http://site4/" ) ),
			ArrayDocumentSummary.parse( testDoc[ 1 ], URI.create( "http://site1/" ) ), ArrayDocumentSummary.parse( testDoc[ 3 ], URI.create( "http://site3/" ) ),
			ArrayDocumentSummary.parse( testDoc[ 2 ], URI.create( "http://site2/" ) ), ArrayDocumentSummary.parse( testDoc[ 0 ], URI.create( "http://site0/" ) ),
			ArrayDocumentSummary.parse( testDoc[ 5 ], URI.create( "http://site5/" ) ) };

	public static ArrayDocumentSummary[] testCollection2;

	public static int[] frequency1;

	public static int[] frequency2;

	public static int numberOfWords1;

	public static int numberOfWords2;

	public static boolean notTearDown;

	static {
		TestCollection.testCollection2 = parseSummaries( testDoc );
		TestCollection.testParse = parseSummaries( testParseDoc );
		TestCollection.testBiParse[0] = parseSummaries( testBiParseDoc[0] );
		TestCollection.testBiParse[1] = parseSummaries( testBiParseDoc[1] );
		TestCollection.frequency1 = makeFrequency( testCollection1 );
		TestCollection.frequency2 = makeFrequency( testCollection2 );
	}

	public static ArrayDocumentSummary[] parseSummaries( String[] docs ) {
		final ArrayDocumentSummary[] r = new ArrayDocumentSummary[ docs.length ];

		for ( int i = 0; i < docs.length; i++ )
			r[ i ] = ArrayDocumentSummary.parse( docs[ i ], null );

		return r;
	}


	public static void pseudoPermute( int[] x, int randSeed ) {
		final Random rng = new Random( randSeed );

		for ( int i = 0; i < x.length - 1; i++ ) {
			final int j = rng.nextInt( x.length - i ) + i;

			x[ i ] ^= x[ j ];
			x[ j ] ^= x[ i ];
			x[ i ] ^= x[ j ];
		}
	}


	public static int[] makeFrequency( ArrayDocumentSummary[] collection ) {
		final Int2IntAVLTreeMap tf = new Int2IntAVLTreeMap();

		tf.defaultReturnValue( 0 );

		for ( ArrayDocumentSummary summary : collection )
			for ( int i = 0; i < summary.size(); i++ ) {
				final int term = summary.term( i );
				tf.put( term, tf.get( term ) + 1 );
			}

		final int[] v = new int[ tf.lastIntKey() + 1 ];

		for ( int term : tf.keySet() )
			v[ term ] = tf.get( term );

		return v;
	}

	public static ArrayArchive makeArrayArchive( ArrayDocumentSummary[] summaries ) throws Exception {
		final ArrayArchive archive = new ArrayArchive();

		for ( ArrayDocumentSummary s : summaries )
			archive.append( s );

		archive.close();

		return archive;

	}

	public static String buildDocumentSequence( String[] sources ) throws Exception {
		String basename = File.createTempFile( "docseq", null ).getAbsolutePath();

		BinIO.storeObject( new InMemoryDocumentCollection( sources ), basename );

		return basename;
	}



	public static String buildSequentialArchive( ArrayDocumentSummary[] summaries ) throws Exception {

		String basename = File.createTempFile( "seqarch", null ).getAbsolutePath();

		final BitstreamArchiveWriter<ArrayDocumentSummary> writer = new BitstreamArchiveWriter<ArrayDocumentSummary>( basename, frequency1, false, CompressionFlags.DEFAULT );

		for ( ArrayDocumentSummary s : summaries ) {
			writer.append( s );

		}

		writer.close();

		return basename;

	}

	/** Creates the archive associated with {@link #testCollection1}, returns the basename. */
	public static String buildUsualSequentialArchive( Map<Component, Coding> codings ) throws Exception {
		String basename = File.createTempFile( "seqarch", null ).getAbsolutePath();

		final BitstreamArchiveWriter<ArrayDocumentSummary> writer = new BitstreamArchiveWriter<ArrayDocumentSummary>( basename, frequency1, false, codings );

		numberOfWords1 = 0;
		for ( ArrayDocumentSummary s : testCollection1 ) {
			writer.append( s );
			numberOfWords1 += s.length();
		}

		writer.close();

		return basename;
	}


	public static String buildTermlistFile() throws Exception {

		File file = File.createTempFile( "tlist", null );
		final PrintWriter pw = new PrintWriter( new OutputStreamWriter( new FileOutputStream( file ) ) );

		for ( String s : testTerms )
			pw.println( s );
		pw.flush();
		pw.close();

		return file.getAbsolutePath();
	}

	/** Creates the sorted archive associated with {@link #testCollection2}. */

	public static String buildRandomAccessArchive( Map<Component, Coding> codings ) throws Exception {
		String basename = File.createTempFile( "randarch", null ).getAbsolutePath();

		final BitstreamArchiveWriter<ArrayDocumentSummary> writer = new BitstreamArchiveWriter<ArrayDocumentSummary>( basename, frequency2, true, codings );

		numberOfWords2 = 0;
		for ( ArrayDocumentSummary s : testCollection2 ) {
			writer.append( s );
			numberOfWords2 += s.length();
		}

		writer.close();

		ComputeArchiveData.run( basename );
		
		return basename;
	}



	/** Cleans up files for any kind of bitstream archive created by this class. */
	public static void cleanupBitstreamArchive( String basename ) throws Exception {
		for ( String i : extensions ) new File( basename + i ).delete();
		new File( basename ).delete();
	}
	
	public static TermProcessor getNullTermProcessor() {
		return NullTermProcessor.getInstance();
	}
	
	public static void compareSummaries( ArrayDocumentSummary[] truth, Archive<? extends DocumentSummary> onDisk ) {
		Iterator<? extends DocumentSummary> onDiskIterator = onDisk.iterator();
		
		for ( int i = 0; i < truth.length ; i ++ )
		{
			TestCase.assertTrue ( onDiskIterator.hasNext() );
			TestCase.assertEquals( truth[i], onDiskIterator.next() );
		}
	}
	
	public static void compareArchives( Archive<? extends DocumentSummary> truth, Archive<? extends DocumentSummary> onDisk  ) {
		TestCase.assertEquals( truth.numberOfWords(), onDisk.numberOfWords());
		TestCase.assertEquals( truth.numberOfDocuments(), onDisk.numberOfDocuments());
		TestCase.assertEquals( truth.numberOfTerms(), onDisk.numberOfTerms() ) ;
		
		for ( int i = 0; i < truth.numberOfTerms(); i++ )
			TestCase.assertEquals ( truth.frequency(i), onDisk.frequency(i));
		
		Iterator<? extends DocumentSummary> truthIterator = truth.iterator();
		Iterator<? extends DocumentSummary> onDiskIterator = onDisk.iterator();
		
		while ( truthIterator.hasNext() )
		{
			TestCase.assertTrue ( onDiskIterator.hasNext() );
			TestCase.assertEquals( truthIterator.next(), onDiskIterator.next() );
		}
	}
}
